/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client.dataobject;

import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.pua.prot.aggregations.Aggregation;
import de.bsvrz.pua.prot.util.ProcessingInformation;

import java.util.*;

/**
 * Repräsentiert die Beschreibung der Aggregationen, die in einem Ergebnisdatensatz vorkommen
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolAggregationData implements ProtocolResult {

	private static final long serialVersionUID = 4008841926493179453L;

	private final List<List<String>> _aggregations;

	private final List<ProcessingInformation.ApplyAggregation> _orders;

	/** Id mit dem Typ des Datums */
	public long[] columnDataType;

	/**
	 * Zeigt an, welche Spalten beim Aggregieren zusammengefasst wurden. Alle Einträge mit der gleichen Nummer gehören zusammen. Einträge mit dem Wert
	 * <code>-1</code> bedeuten, dass diese Spalte mit keiner anderen Spalte zusammengefasst wurde. Ist <code>null</code>, wenn die Aggregationen nicht nach
	 * 'spalte', sondern nur nach 'objekt' durchgeführt wurden.
	 */
	public int[] linkedAttributes;

	/** @param size Anzahl der Einträge.
	 * @param orders Orders*/
	public ProtocolAggregationData(int size, final List<ProcessingInformation.ApplyAggregation> orders) {
		_orders = orders;
		_aggregations = new ArrayList<List<String>>();
		columnDataType = new long[size];
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof ProtocolAggregationData) {
			ProtocolAggregationData other = (ProtocolAggregationData)object;
			return Arrays.equals(this.columnDataType, other.columnDataType) && this.getAggregations().equals(other.getAggregations());
		}
		else {
			return false;
		}
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		
		return _aggregations.toString();
	}

	@Override
	public String toString(final DataModel dataModel) {
		StringBuilder sb = new StringBuilder();
		sb.append(String.format("%-28s\t", "Aggregationen"));
		for(List<String> c : getAggregations()) {
			sb.append(String.format("%28s\t", c.toString()));
		}
		return sb.toString();
	}

	/** @see de.bsvrz.pua.prot.client.dataobject.ProtocolResult#getType() */
	public int getType() {
		return ProtocolResult.AGGREGATION_DATA;
	}

	/**
	 * Das Listenelement entspricht dem Namen der Aggregationen die auf die Spalte angewandt wurden. Leere Liste falls keine Aggregation verwendet wurde.
	 *
	 * @see de.bsvrz.pua.prot.aggregations.Aggregation#getAggregationName()
	 * @return Liste
	 */
	public List<List<String>> getAggregations() {
		return Collections.unmodifiableList(_aggregations);
	}

	public void addAggregations(final List<Class<? extends Aggregation>> aggregationClasses) {
		final ArrayList<String> list = new ArrayList<String>(aggregationClasses.size());
		for(final Class<? extends Aggregation> aggregationClass : aggregationClasses) {
			try {
				list.add(aggregationClass.newInstance().getAggregationName());
			}
			catch(Exception ignored) {
				list.add(aggregationClass.getName());
			}
		}
		_aggregations.add(list);
	}

	public List<ProcessingInformation.ApplyAggregation> getOrders() {
		return Collections.unmodifiableList(_orders);
	}
}
