/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing.archivebuffer;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.archive.*;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.processing.ProcessingBuffer;
import de.bsvrz.pua.prot.processing.ProcessingInterface;
import de.bsvrz.pua.prot.processing.ValueProvider;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.IntermediateDataSet;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.attributes.AttributeGroupDescription;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealElement;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.ArrayList;
import java.util.List;

/**
 * Puffer für die Daten vom Archivsystem, die von der Datenaufbereitung angefordert werden. Der folgende Algorithmus wird für jeden Zeitbereich wiederholt:<br>
 * Für jedes reale Element, d.h. für jedes reale Attribut und für jede Attributgruppe wird eine Archivanfrage über den Zeitbereich und Datenart gestellt. Der
 * jeweils erste pro Element erhaltene Datensatz wird markiert. Anschließend werden die Datensätze verschränkt. Dabei wird ein "Gewinner" festgestellt. Der
 * Gewinner ist das Element mit dem kleinsten Datenzeitstempel. Danach werden die Werte der übrigen Attribute aufgefüllt und als Ausgangsdatensatz in einem
 * Puffer abgelegt. Im nächsten Schritt wird ein weiterer Datensatz des Gewinners abgerufen, markiert mit den Datensätzen der übrigen Elementen verschränkt und
 * dabei ein neuer "Gewinner" bestimmt(*). Nun wird die Datenaufbereitung benachrichtigt, dass ein Ausgangsdatensatz vorliegt. Grund für diese Verzögerung ist,
 * dass die Bearbeitung eines Ausgangsdatensatzes erst vollständig abgeschlossen ist, wenn der nächste Ausgangsdatensatz erzeugt wurde. Der in (*) gewonnene
 * Ausgangsdatensatz wird im Puffer abgelegt, und der Algorithmus wiederholt sich bis alle Daten vom Archivsystem abgerufen wurden. Ausgangsdatensatz wird
 * ebenfalls in einem Puffer abgelegt. Wird in einem eigenen Thread gestartet, da ggf. auf die Antwort des Archivsystems gewartet werden muss. Da Archivdaten
 * streambasiert abgefragt werden können, stellt der ArchivBuffer sicher, dass sich im Puffer für die Ausgangsdatensätze nur eine bestimmte Anzahl von Einträgen
 * ansammelt. Wird dieses Limit ({@link #MAX_THRESHOLD}) erreicht, so stellt der Archivbuffer die arbeit ein, bis sich der Puffer wieder fast vollständig
 * ({@link #MIN_THRESHOLD}) geleert hat.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ArchiveBuffer extends ProcessingBuffer {

	/**
	 * Anfragepriorität an das Archivsystem.
	 *
	 * @see ArchiveRequestManager#request(de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority,de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification)
	 */
	public static final ArchiveQueryPriority PRIORITY = ArchiveQueryPriority.LOW;

	/** Anzahl Ausgangsdatensätze, die im Ausgangspuffer liegen. */
	public static int MAX_THRESHOLD = 101;

	/** Anzahl Ausgangsdatensätze, die mindestens im Ausgangspuffer liegen sollten. */
	public static int MIN_THRESHOLD = 10;

	/** Zeigt an ob die Datensammlung abgeschlossen ist. */
	private boolean doneCollecting = false;

	/** Parameter der Archivanfrage */
	private ArchiveUser _archiveUser;

	/** Objekt das die Aggregierung durchführt */
	private Aggregator _aggregator;

	/** Index des aktuellen Zeitintervalls. */
	private int _currentPeriod;

	/** Index des aktuellen Zeitintervalls. Anders als _currentPeriod wird dies erst später Aktualisert, damit aggregationen use. noch das aktuelle Intervall benutzen */
	private int _currentDisplayPeriod;

	/** <code>true</code> falls die Intervall-Aggregationsdatensätze erstellt werden müssen. Dies ist nach jedem Intervall der Fall. */
	private boolean _endOfInterval = false;

	/** <code>true</code> falls die Gesamt-Aggregationsdatensätze erstellt werden müssen. Dies ist am Ende der Protokollerstellung der Fall. */
	private boolean _endOfProtocol = false;

	/**
	 * Startet den Online-ProcessingBuffer. Meldet sich für jede benötigten Datenidentifikation beim Datenverteiler an. {@link #init()} sollte unverzüglich nach
	 * Erzeugen des Objekts aufgerufen werden.
	 *
	 * @param dav        Verbindung zum Datenverteiler
	 * @param processor  Objekt das die Datenaufbereitung durchführt. Wird jedes Mal benachrichtigt, wenn ein Ausgangsdatensatz vorliegt.
	 * @param configAuth Konfigurationsverantwortlicher, dessen Archivsystem verwendet wird.
	 * @param pi         Informationen zur Datenaufbereitung. Die Zeitbereiche müssen bereits sortiert und zusammengefasst sein!
	 * @param periods    Zeitbereiche, in denen der Archivbuffer Daten sammeln soll. Inhalt wird geändert!
	 *
	 * @param scriptName Bezeichnung des Skripts zur Verwendung in Fehlerausgaben.
	 * @throws FailureException Fehler bei der Kommunikation mit der Konfiguration
	 */
	public ArchiveBuffer(
			ClientDavInterface dav, ProcessingInterface processor, ConfigurationObject configAuth, ProcessingInformation pi, List<Tuple<Long, Long>> periods, final String scriptName)
			throws FailureException {
		super(dav, processor, pi, periods, 0, scriptName);

		// Aggregationen aufbauen:
		_aggregator = new Aggregator(pi);

		// Archivanfrage festlegen
		if(periods.isEmpty()) {
			doneCollecting = true;
			setDone();
		}
		else {
			_archiveUser = createArchiveUser(configAuth);
			Tuple<Long, Long> period = periods.get(_currentPeriod);
			_archiveUser.setTimeSpec(new ArchiveTimeSpecification(TimingType.DATA_TIME, false, period.first, period.last));

			_archiveUser.setArchiveDataKinds(pi.getArchiveDataKinds());
		}
	}

	/**
	 * Erzeugt eine Verbindung zum Archivsystem die zum Absenden von Daten mit Standardparametern verwendet werden kann.Die Standardparameter sind:<br> -
	 * Sortierung der nachgelieferten Daten nach Datenzeit<br> - Delta oder Standardanfrage je nach {@link ProcessingInformation#getProtocolType()}
	 *
	 * @param configAuth Konfigurationsverantwortlicher, dessen Archivsystem zu verwenen ist.
	 *
	 * @return Anfrageschnittstelle zum Archivsystem
	 */
	private ArchiveUser createArchiveUser(SystemObject configAuth) {
		// Bestimmte Parameter sind für alle Anfragen gleich:
		ArchiveOrder sortOrder = ArchiveOrder.BY_DATA_TIME;
		ArchiveRequestOption requestOption;
		if(_processingInformation.getProtocolType() == ProtocolType.DeltaProtocol) {
			requestOption = ArchiveRequestOption.DELTA;
		}
		else {
			requestOption = ArchiveRequestOption.NORMAL;
		}

		return new ArchiveUser(_connection.getArchive(configAuth), null, null, sortOrder, requestOption);
	}

	/**
	 * Sendet erste Anfragen an das Archivsystem. Startet Anschließend einen Thread, der die Daten vom Archivsystem entgegennimmt. Falls es zu einem Fehler bei der
	 * Archivanfrage kommt, wird der Thread nicht gestartet, und {@link ProcessingBuffer#abort()} aufgerufen.
	 *
	 * @throws FailureException     Fehler bei der Archivanfrage.
	 * @throws InterruptedException Warten auf Archivantwort wurde unterbrochen.
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#init()
	 */
	@Override
	public void init() throws FailureException, InterruptedException {
		// OPT Anfragen optimieren: Nur eine Anfrage je DI!
		try {
			if(!doneCollecting) {
				// Reale Attribute
				boolean aggregationExists = false;
				for(RealElement realElement : _processingInformation.getAllRealElements()) {
					if(realElement instanceof RealAttributeDescription) {
						RealAttributeDescription re = (RealAttributeDescription)realElement;
						if(re.qualifier.object != null){
							RealArchiveAttribute raa = new RealArchiveAttribute(_archiveUser, re, _processingInformation.getProtocolType(), _processingInformation.isShown(realElement));
							raa.request();
							_realElements.put(re, raa);
						}
						aggregationExists |= re.getAggregationClasses().size() > 0;
					}
					else if(realElement instanceof AttributeGroupDescription) {
						AttributeGroupDescription re = (AttributeGroupDescription)realElement;
						if(re.qualifier.object != null){
							RealArchiveAttributeGroup raa = new RealArchiveAttributeGroup(_archiveUser, re, _processingInformation.getProtocolType(), _processingInformation.isShown(realElement));
							raa.request();
							_realElements.put(re, raa);
						}
						aggregationExists |= re.getAggregationClasses().size() > 0;
					}
				}
				setTempAttributes();

				_debug.info("Stelle Archivanfragen für " + this.getName());
				_archiveUser.execute();
				this.setAggregate(aggregationExists);
				this.start();
			}
		}
		catch(FailureException e) {
			this.abort();
			throw e;
		}
		catch(InterruptedException e) {
			this.abort();
			throw e;
		}
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#applyAggregations(List)  */
	@Override
	public boolean applyAggregations(List<BaseDataSet> baseData) {
		return _aggregator.aggregate(baseData);
	}

	/**
	 * Fragt Daten für das nächste Intervall an. Benachrichtigt außerdem die Aggregationen, wenn das Intervall abgeschlossen wurde.
	 *
	 * @return <code>true</code> wenn Daten für das nächste Intervall angefragt werden. <code>false</code> sonst, z.B. weil keine weiteren Zeitintervalle
	 *         existieren
	 *
	 * @throws FailureException     Fehler bei der Archivanfrage
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	private boolean requestNextInterval() throws FailureException, InterruptedException {
		/*
		 * Flag setzen, dass beim nächsten Aufruf von storeAggregatedData ein Interval-Aggregationsdatensatz erstellt
		 * werden soll.
		 */
		_endOfInterval = true;

		long endOfCurrentInterval = _periods.get(_currentPeriod).last;
		_currentPeriod++;
		if(_currentPeriod == _periods.size()) // Letzten Zeitbereich beendet.
		{
			// Flag setzen, dass die Gesamt-Aggregationsdatensätze erstellt werden müssen
			_endOfProtocol = true;
			this.nextInterval(endOfCurrentInterval);
			return false;
		}
		else { // Es sind noch weitere Zeitbereiche abzufragen
			Tuple<Long, Long> period = _periods.get(_currentPeriod);
			ArchiveTimeSpecification timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false, period.first, period.last);
			_archiveUser.setTimeSpec(timeSpec);
			requestData();

			this.nextInterval(endOfCurrentInterval);
			return true;
		}
	}

	/**
	 * Wendet den Nachfilter an und speichert den Ausgangsdatensatz im Puffer ab. Der Datenstaz wird nur gespeichert, falls er auch nach der Anwendung des
	 * Nachfilters nicht <code> null</code> ist
	 *
	 * @param ids Ausgangsdatensatz. Darf <code>null</code> sein: Dann wird der Datensatz nicht gespeichert.
	 */
	private void applyPostFilterForAggregatedData(IntermediateDataSet ids) {
		if(ids != null) {
			if(_processingInformation.getUserFilter() != null) ids = _processingInformation.getUserFilter().processAggregationRecord(ids);
			if(ids != null) _buffer.push(ids);
		}
	}

          	/**
	 * Stellt Archivanfragen für alle realen Elemente mit den momentan engetragenen Werten von {@link #_archiveUser}
	 *
	 * @throws FailureException     Fehler bei der Archivanfrage.
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	protected void requestData() throws FailureException, InterruptedException {
		// OPT Umschreiben, so dass nur ein Request für ArchiveUser nötig
		for(ValueProvider vp : _realElements.values()) {
			((RealArchiveElement)vp).request();
		}
		_archiveUser.execute();
	}

	/**
	 * Überprüft ob die Datensammlung abgeschlossen ist.
	 *
	 * @return <code>True</code> falls die Datensammlung abgeschlossen ist. Überprüft zudem, wieviele Elemente sich in der Warteschlange befinden. Ist die
	 *         Warteschlange bereits über {@link #MAX_THRESHOLD} gefüllt, so wird der ArchivBuffer angehalten, bis die Queue wieder fast vollständig ({@link
	 *         #MIN_THRESHOLD}) geelert ist.
	 *
	 * @throws FailureException     Fehler bei der Archivanfrage. Es wird eine Archivanfrage gestellt, wenn Daten für ein weiteres Intervall angefragt werden.
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 * @see de.bsvrz.pua.prot.processing.ProcessingBuffer#isDoneCollecting()
	 */
	@Override
	protected boolean isDoneCollecting() throws FailureException, InterruptedException {
		if(doneCollecting || isAbort()) {
			return true;
		}
		else {
			if(_buffer.size() > MAX_THRESHOLD) {
				synchronized(this) {
					while(_buffer.size() >= MIN_THRESHOLD) {
						this.wait();
					}
				}
			}
			boolean temp = false;
			for(ValueProvider vp : _realElements.values()) {
				temp |= vp.hasNext();
			}
			if(!temp) // falls keine weiteren Daten abzuholen sind: prüfen, ob noch eine weitere Archivanfrage
			// gestellt wird
			{
				if(!requestNextInterval()) {
					doneCollecting = true;
					return true;
				}
				else {
					return false;
				}
			}
			else {
				return false;
			}
		}
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#isDone() */
	@Override
	public boolean isDone() {
		return _done || (doneCollecting && _buffer.isEmpty() && _dataSetBuilder.isDone());
	}

	/**
	 * Zeigt an ob Daten abgeholt werden können.
	 *
	 * @return <code>true</code> falls Daten mittels {@link #take()} abgeholt werden können.
	 */
	@Override
    public boolean hasData() {
		return (_buffer.size() >= 2 || (doneCollecting && !_buffer.isEmpty()));
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#take() */
	@Override
	public IntermediateDataSet take() throws InterruptedException {
		synchronized(this) {
			if(_buffer.size() < MIN_THRESHOLD){
				notifyAll();
			}
		}
		return super.take();
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#isListAggregation() */
	@Override
	protected boolean isListAggregation() {
		return _aggregator.isListAggregation();
	}

	/**
	 * Zeigt an welche Attribute durch die Aggregation <code>spalte</code> zusammengefasst werden. Zusammengehörige Spalten werden duch die gleichen Nummern
	 * gekennzeichnet.
	 *
	 * @return Zusammen gehörende Spalten. <code>null</code> falls die <code>spalten</code> Aggregation nicht verwendet wird.
	 */
	public int[] getLinkedAttributes() {
		return _aggregator.getLinkedAttributes();
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#storeAggregatedData(byte) */
	@Override
	public void storeAggregatedData(byte status) {
		if(_endOfInterval) {
			{
				_endOfInterval = false;
				// Intervall Aggregation auch für das letzte Intervall durchführen
				final IntermediateDataSet[] intermediateData = _aggregator.endOfInterval(status, _currentDisplayPeriod);
				applyPostFilterForAggregatedData(intermediateData[0]);
				applyPostFilterForAggregatedData(intermediateData[1]);
			}

			if(_endOfProtocol) {
				_endOfProtocol = false;
				// Gesamt - Aggregation
				final IntermediateDataSet[] intermediateData = _aggregator.endOfProtocol(_currentDisplayPeriod);
				applyPostFilterForAggregatedData(intermediateData[0]);
				applyPostFilterForAggregatedData(intermediateData[1]);
			}

			_currentDisplayPeriod++;
		}
	}


	/** @see de.bsvrz.pua.prot.processing.ProcessingBuffer#processNewData(ValueProvider[], ArrayList)
	 */
	@Override
	protected void processNewData(ValueProvider[] elements, ArrayList<ValueProvider> winners) throws FailureException, InterruptedException {
		// Werte holen
		if(winners.isEmpty()) winners.addAll(_realElements.values());
		for(ValueProvider winner : winners) {
			winner.take(0);
			winner.updateDataInformation(); // Markierungsinformation updaten
		}

		/*
		 * >0 falls Gewinner im Spaltenbereich, <0 falls Gewinner woanders, 0 falls kein // Gewinner gefunden
		 */
		int hasWinners = 0;
		// Element mit minimalem DZS auswählen
		hasWinners = getWinners(elements, winners);
		if(hasWinners == 0) // ist true falls keines der bisher ausgewählten Elemente weitere Daten hat.
		{
			winners.addAll(_realElements.values());
			hasWinners = getWinners(elements, winners);
		}
		else if(hasWinners > 0) // Im Spaltenbereich hat sich was getan -> Daten aufbereiten
		{
			// Minimalen Zeitstempel merken den nach oben ziehen...
			long winnerTimeStamp = winners.get(0).getMostRecentTimeStamp();

			_dataSetBuilder.collectData(getInsertEmpty(), winnerTimeStamp);
			_dataSetBuilder.createIntermediateDataSet(false);
		}
	}
}
